#ifdef __APPLE__
#import <AvailabilityMacros.h>
#else
#include <math.h>
#ifndef NAN
#define NAN FP_NAN
#endif
#endif

#import <Foundation/Foundation.h>

#import <stdio.h>
#import <math.h>
#import <assert.h>
#import <string.h>

#import <vector.h>
#import <camera.h>

//---- @implementation section ----

@implementation camera_t

- (id) init
{
  if(!(self = [super init]))
    return nil;
  cookie = CAM_COOKIE;
  name = [[NSString alloc] init];
  view_point = [[vec_t alloc] init];
  return self;
}

- (NSString *) setname: (NSString *) _name
{
  /* C++-style
  if(name != _name) {
    // release old obj
    [name release];
    name = [_name retain];
  }
  */
  /* Objective C style */
  // release old obj but only after we retain new one (ok if name == _name)
  [name autorelease];
  name = [_name retain];
}

- (NSString *) getname { return name; }
- (int) getpw { return pixel_dim[0]; }
- (int) getph { return pixel_dim[1]; }
- (double) getww { return world_dim[0]; }
- (double) getwh { return world_dim[1]; }
- (vec_t *) view_point { return view_point; }

  // operators

  // members

- (void) read: (FILE *) _in
{
	char	c, _name[256];
	int	count;

  // assume file pointer has already consumed 'camera' token
  assert((count = fscanf(_in,"%s",_name))==1);
  [self setname: [[NSString alloc] initWithUTF8String: _name]];

  // consume all chars until we get to '{'
  while((c=fgetc(_in)) != EOF && c != '{');

  // loop until we hit '}'
  while((c=fgetc(_in)) != EOF && c != '}') {
    // put character back into input stream
    fputc(c,_in);

    assert((count = fscanf(_in,"%s",_name))==1);

    if(!strcmp(_name,"pixeldim"))
      fscanf(_in,"%d %d",&pixel_dim[0],&pixel_dim[1]);
    if(!strcmp(_name,"worlddim"))
      fscanf(_in,"%lf %lf",&world_dim[0],&world_dim[1]);
    if(!strcmp(_name,"viewpoint")) [view_point read: _in];

    // consume whitespace
    while((c=fgetc(_in)) != EOF && c != '\n');
  }
}

- (void) write: (FILE *) _out
{
  assert(cookie == CAM_COOKIE);
 
  fprintf(_out,"camera %s\n",[name UTF8String]);
  fprintf(_out,"{\n");
  fprintf(_out,"   pixeldim %d %d\n",pixel_dim[0],pixel_dim[1]);
  fprintf(_out,"   worlddim %3.0lf %3.0lf\n",world_dim[0],world_dim[1]);
  [view_point write: _out: "   viewpoint"];
  fprintf(_out,"}\n\n");
}

@end
