#import <Foundation/Foundation.h>

#import <stdio.h>
#import <stdlib.h>
#import <string.h>
#import <assert.h>
#import <math.h>

#import <timer.h>
#import <vector.h>
#import <pixel.h>
#import <list.h>
#import <camera.h>
#import <material.h>
#import <object.h>
#import <plane.h>
#import <model.h>

//---- prototype section ----
pixel_t *ray_trace(model_t *,vec_t *,vec_t *,double);
int main(int argc, const char *argv[]);

//---- program section ----

int main(int argc, const char *argv[])
{
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

	atimer_t	*timer = [(atimer_t *)[atimer_t alloc] init];

	model_t		*model = [(model_t *)[model_t alloc] init];
	FILE		*model_file=NULL;

	int		ncores=1;

  NSLog(@"main start");

  if(argc != 2) {
    fprintf(stderr,"Usage: %s <model_file>\n",argv[0]);
    exit(1);
  }

  // open model file
  if((model_file = fopen(argv[1],"r")) == NULL) {
     printf("Error opening input file: %s\n",argv[1]);
     return;
  }

  // read in model
  [model read: model_file];

  // close file
  fclose(model_file);

  // verify that we have the model
  [model write: stderr];

	int		i,x,y;
	int		w = [model getpw], h = [model getph];
	double		wx,wy,wz=0.0;
	double		ww = [model getww], wh = [model getwh];
	vec_t		*pos = [[[model cam] view_point] copy];
	vec_t		*dir = [(vec_t *)[vec_t alloc] init];
	vec_t		*pix = [(vec_t *)[vec_t alloc] init];
	pixel_t		*color = [(pixel_t *)[pixel_t alloc] init: 0.0:0.0:0.0];
	irgb_t		*imgloc,*img=NULL;

  // allocate 2D image array and zero out
  assert((img = (irgb_t *)malloc(w*h*sizeof(irgb_t)))!=NULL);
  memset(img,0,w*h*sizeof(irgb_t));

  [timer start];
  for(y=h-1;y>=0;y--) {
    for(x=0;x<w;x++) {
      wx = (double)x/(double)(w-1) * ww;
      wy = (double)y/(double)(h-1) * wh;

      // set pixel position vector (in world coordinates)
      [pix set: wx: wy: wz];

      // compute the vector difference
      dir = [pix minus: pos];

      // normalize dir
      dir = [[dir unit] retain];

      // trace ray (autoreleases)
      color = ray_trace(model,pos,dir,0.0);

      // where are we in the image (using old i*c + j)
      imgloc = img + y*w + x;
      for(i=0;i<3;i++) (*imgloc)[i] = [color get: i];

    }
  }
  [dir release];
  [pix release];
  [pos release];

  [timer stop];

  fprintf(stderr,"cores: %d\n",ncores);
  fprintf(stderr,"time: %.2lf ms\n",[timer elapsed_ms]);

  fprintf(stdout,"P6 %d %d 255\n",w,h);
  for(y=h-1;y>=0;y--) {
    for(x=0;x<w;x++) {
      imgloc = img + y*w + x;
      fwrite(imgloc,sizeof(irgb_t),1,stdout);
    }
  }

  [pool drain];

  return 0;
}

pixel_t *ray_trace(model_t  *model,
                   vec_t    *pos,
                   vec_t    *dir,
                   double    raydist)
{
//	pixel_t	 *color = [(pixel_t *)[pixel_t alloc] init: 0.0: 0.0: 0.0];
	pixel_t	 *color = nil;

        double	  dis=0.0;
	vec_t	 *hit = [(vec_t *)[vec_t alloc] init: 0.0: 0.0: 0.0];
	vec_t	 *N = [(vec_t *)[vec_t alloc] init: 0.0: 0.0: 0.0];
        object_t *obj=NULL;
        pixel_t  *ambient=nil,*diffuse=nil,*specular=nil;

  // get closest object, if any
  if(!(obj = [model find_closest: pos: dir: &dis: &hit: &N]))
    return;

  // if hit, print info
  if(dis > 0) {
    // accumulate distance travelled by ray
    raydist += dis;

    // get object material properties
    ambient = [[obj mat] getambn];
    diffuse = [[obj mat] getdiff];
    specular = [[obj mat] getspec];

    // for now just use ambient color
    [color autorelease];
    color = [ambient copy];

    // scale by distance (autoreleases)
    [color autorelease];
    color = [color scale: 1.0/raydist];
  }

  // scale to pixel values (autoreleases)
  return [color scale: 255.0];
}
