#ifdef __APPLE__
#import <AvailabilityMacros.h>
#else
#include <math.h>
#ifndef NAN
#define NAN FP_NAN
#endif
#endif

#import <Foundation/Foundation.h>
#import <Foundation/NSString.h>

#import <stdio.h>
#import <string.h>
#import <math.h>

#import <vector.h>
#import <pixel.h>
#import <list.h>
#import <camera.h>
#import <material.h>
#import <object.h>
#import <plane.h>
#import <model.h>

//---- @implementation section ----

@implementation model_t

- (id) init
{
  if(!(self = [super init]))
    return nil;
  cam  = [(camera_t *)[camera_t alloc] init];
  mats = [(list_t *)[list_t alloc] init];
  objs = [(list_t *)[list_t alloc] init];
  return self;
}

- (camera_t *) cam { return cam; }
- (list_t *) mats { return mats; }
- (list_t *) objs { return objs; }
- (int) getpw { return [cam getpw]; }
- (int) getph { return [cam getph]; }
- (double) getww { return [cam getww]; }
- (double) getwh { return [cam getwh]; }

- (void) read: (FILE *) _in
{
        char             token[16];
        material_t      *mat = nil;
        object_t        *obj = nil;
        NSMutableString *str = nil;

  while(fscanf(_in,"%s",token)==1) {

    if(!strcmp(token,"camera")) {
      [cam read: _in];
    }
 
    if(!strcmp(token,"material")) {
      mat = [(material_t *)[material_t alloc] init];
      [mat read: _in];
      [mats add: mat];
    }

    if(!strcmp(token,"plane")) {
      obj = [(plane_t *)[plane_t alloc] init];
      str = [[NSMutableString alloc] initWithUTF8String: token];
      [obj settype: str];
      [obj read: _in: mats];
      [objs add: obj];
      [str release];
    }
  }
}

- (void) write: (FILE *) _out
{
  [cam write: _out];

  [mats reset];
  while(![mats end]) {
    [[mats data] write: _out];
    [mats next];
  }

  [objs reset];
  while(![objs end]) {
    [[objs data] write: _out];
    [objs next];
  }
}

- (object_t *) find_closest: (vec_t *) pos: (vec_t *) dir:
                            (double *) dis:
                            (vec_t **) hit: (vec_t **) N
{

	// candidate object (as we iterate), distance, hit point, and N
	double		c_dis;
	vec_t		*c_hit=[(vec_t *)[vec_t alloc] init];
	vec_t		*c_N=[(vec_t *)[vec_t alloc] init];
	object_t	*c_obj=nil;

	// closest object thus far (and related data, incl. hit, N)
	double		closest_dis=INFINITY;
	vec_t		*closest_hit=[(vec_t *)[vec_t alloc] init];
	vec_t		*closest_N=[(vec_t *)[vec_t alloc] init];
	object_t	*closest_obj=nil;

  // check each object: basic list iteration
  for([objs reset]; ![objs end]; [objs next]) {

    // get candidate object
    c_obj = [objs data];

    if(![c_obj conformsToProtocol: @protocol(intersecting)])
      NSLog(@"Warning: object does not support intersecting protocol");

    // if same object or no hit, keep going (side effect: dist calculated)
    if((c_dis = [c_obj hits: pos: dir: &c_hit: &c_N])<0.)
      continue;
    else
    if((0.000001 < c_dis) && (c_dis < closest_dis)) {
      closest_dis = c_dis;
      closest_obj = c_obj;
      [closest_hit autorelease]; closest_hit = [c_hit copy];
      [closest_N autorelease]; closest_N = [c_N copy];
    } else
      continue;
  }

  if(closest_obj) {
    [*hit autorelease]; *hit = [closest_hit copy];
    [*N autorelease]; *N = [closest_N copy];
    (*dis) += closest_dis;
  }

  [c_hit release];
  [c_N release];
  [closest_hit release];
  [closest_N release];

  return closest_obj;
}

@end
