#ifdef __APPLE__
#import <AvailabilityMacros.h>
#else
#include <math.h>
#ifndef NAN
#define NAN FP_NAN
#endif
#endif

#ifndef OBJ_COOKIE
#define OBJ_COOKIE 12345678
#endif

#import <Foundation/Foundation.h>

#import <stdio.h>
#import <math.h>
#import <assert.h>
#import <string.h>

#import <vector.h>
#import <pixel.h>
#import <list.h>
#import <material.h>
#import <object.h>
#import <plane.h>

//---- @implementation section ----

@implementation plane_t

- (id) init
{
  if(!(self = [super init]))
    return nil;
  normal = [[vec_t alloc] init];
  point = [[vec_t alloc] init];
  ndotq = 0.0;
  return self;
}

- (double) hits: (vec_t *) _pos: (vec_t *) _dir: (vec_t **) _hit: (vec_t **) _N
{
	double	ndotd,t,ndotb,precision=0.000001;

  assert(cookie == OBJ_COOKIE);

  // get angle between view direction and the plane normal
  ndotd = [_dir dot: [normal unit]];

  // if ndotd == 0, ray is parallel to the plane, no intersection
  if(fabs(ndotd) < precision) return -1.0;

  // find distance of ray pos to plane: N . b + ndotq
  ndotb = [_pos dot: [normal unit]];

  // here we need to temporarily negate N to find distance along ray to plane
  t = (ndotq - ndotb) / ndotd;

  // if t < 0 then intersection "behnid" where ray started from
  if(t <= 0.0) return -1.0;

  // hit point is obtained by moving along dir by t amount
  [last_hit autorelease]; last_hit = [[[_dir scale: t] plus: _pos] retain];

  // if z > 0 then hit point is in front of image plane: invalid intersection
  if([last_hit get: 2] > 0.0) return -1.0;
  else {
    [*_hit autorelease];  *_hit = [last_hit copy];
    [*_N autorelease];  *_N = [normal copy];
  }

  return t;
}

  // operators

  // members

- (void) read: (FILE *) _in: (list_t *)mats
{
	char             c, _name[256];
	int              count;
	NSMutableString *str;

  [super read: _in: mats];

  // loop until we hit '}'
  while((c=fgetc(_in)) != EOF && c != '}') {
    // put character back into input stream
    fputc(c,_in);

    assert((count = fscanf(_in,"%s",_name))==1);

    if(!strcmp(_name,"normal")) [normal read: _in];
    if(!strcmp(_name,"point")) [point read: _in];

    // consume whitespace
    while((c=fgetc(_in)) != EOF && c != '\n');
  }

  // set plane normal
//[normal autorelease];
//normal = [[normal unit] retain];

  // set object's last normal
  [last_N autorelease];
  last_N = [normal copy];

  // set plane equation (A,B,C,D) with N = (A,B,C) and D = N.point
  ndotq = [point dot: [normal unit]];
}

- (void) write: (FILE *) _out
{
  assert(cookie == OBJ_COOKIE);

  [super write: _out];

  fprintf(_out,"  ");

  [normal write: _out: " normal"];
  [point write: _out: "   point"];

  fprintf(_out,"}\n\n");
}

@end
